import math

from django.conf import settings
from django.utils.functional import cached_property
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.exceptions import NotFound, ValidationError
from rest_framework.generics import CreateAPIView, RetrieveAPIView
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from SolrClient import SolrClient
from SolrClient.exceptions import SolrError

from arkindex.documents.models import EntityType, MetaType
from arkindex.documents.serializers.search import (
    CorpusSearchQuerySerializer,
    CorpusSearchResultSerializer,
    ReindexCorpusSerializer,
)
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Role

solr = SolrClient(settings.SOLR_API_URL)


def parse_facet(name, value):
    # For enums returned by serializer EnumFields, get the string value
    if isinstance(value, (MetaType, EntityType)):
        value = value.value
    # Escape quotes and backslashes to avoid query injection
    value = value.replace("\\", "\\\\").replace('"', '\\"')
    return f'{name}:"{value}"'


@extend_schema_view(
    get=extend_schema(
        operation_id="SearchCorpus",
        tags=["search"],
        parameters=[CorpusSearchQuerySerializer],
        responses=CorpusSearchResultSerializer
    )
)
class CorpusSearch(CorpusACLMixin, RetrieveAPIView):
    """
    Get a list of Solr documents in a specific corpus whose transcription text, metadata value or element and entity name correspond to the given query.
    """

    def build_solr_query(self, query, sources, facets):
        sources_query = " OR ".join(f"{source}_text:({query})" for source in sources)
        facets_query = " AND ".join(parse_facet(name, value) for name, value in facets.items())
        return f"({sources_query}) AND ({facets_query})" if facets_query else sources_query

    def solr_search(self, collection_name, query, sort, page):
        sort_params = {
            "relevance": "score desc",
            "element_name": "element_text asc",
        }
        if sort not in sort_params:
            raise ValueError("Unknown sort parameter")

        return solr.query(collection_name, {
            "q": query,
            "start": (page - 1) * settings.SOLR_PAGINATION_SIZE,
            "sort": sort_params[sort],
            "rows": settings.SOLR_PAGINATION_SIZE,
            "facet": True,
            "facet.field": [
                "element_type",
                "element_worker",
                "transcription_worker",
                "classification_name",
                "classification_worker",
                "metadata_name",
                "metadata_type",
                "metadata_worker",
                "entity_type",
                "entity_worker"
            ]
        })

    def build_pagination(self, url, page, nb_results):
        nb_page = math.ceil(nb_results / settings.SOLR_PAGINATION_SIZE)
        previous_url = replace_query_param(url, "page", page - 1) if page > 1 else None
        next_url = replace_query_param(url, "page", page + 1) if page < nb_page else None
        return previous_url, next_url

    def get(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES["search"]:
            raise ValidationError(["Search features are not available on this instance."])

        corpus = self.get_corpus(kwargs["pk"])
        collection_name = f"project-{corpus.id}"

        if not corpus.indexable:
            raise ValidationError([f"Corpus {corpus.id} is not indexable."])

        if not solr.collections.exists(collection_name):
            raise NotFound(f"Corpus index {collection_name} not found.")

        # Serialize params
        params = request.query_params.dict()
        # Convert the sources string into a valid list
        # Support both possible urls
        if "sources" in params or "sources[]" in params:
            params["sources"] = request.query_params.getlist("sources") + request.query_params.getlist("sources[]")
        serializer = CorpusSearchQuerySerializer(data=params)
        serializer.is_valid(raise_exception=True)
        q, sources, only_facets, sort, page = map(
            lambda key: serializer.validated_data.pop(key),
            ("query", "sources", "only_facets", "sort", "page")
        )

        # Paginated Solr search
        query = self.build_solr_query(q, sources, serializer.validated_data)
        try:
            results = self.solr_search(collection_name, query, sort, page)
        except SolrError:
            raise ValidationError([f'Query "{q}" is not valid.'])

        # Build response
        nb_results = results.get_num_found()
        previous_url, next_url = self.build_pagination(request.build_absolute_uri(), page, nb_results)

        return Response(CorpusSearchResultSerializer({
            "count": nb_results if not only_facets else 0,
            "number": page,
            "next": next_url,
            "previous": previous_url,
            "results": results.docs if not only_facets else None,
            "facets": results.get_facets()
        }).data, status=status.HTTP_200_OK)


@extend_schema_view(post=extend_schema(operation_id="BuildSearchIndex", tags=["search"]))
class SearchIndexBuild(CorpusACLMixin, CreateAPIView):
    """
    Starts an indexation task for a specific corpus, ran by RQ
    """
    permission_classes = (IsVerified, )
    serializer_class = ReindexCorpusSerializer

    @cached_property
    def corpus(self):
        corpus = self.get_corpus(self.kwargs["pk"], role=Role.Admin)
        if not corpus.indexable:
            raise ValidationError({"__all__": ["This project is not indexable."]})
        if not corpus.types.filter(indexable=True).exists():
            raise ValidationError({"__all__": ["There are no indexable element types for this project."]})
        return corpus

    def get_serializer_context(self):
        return {
            **super().get_serializer_context(),
            "corpus_id": self.corpus.id,
            "user_id": self.request.user.id,
        }

    def create(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES["search"]:
            raise ValidationError({
                "__all__": ["Building search index is unavailable due to the search feature being disabled."]
            })
        return super().create(request, *args, **kwargs)
