from datetime import datetime

from enumfields import Enum


class DateType(Enum):
    Exact = "exact"
    Lower = "lower"
    Upper = "upper"
    Unknown = "unknown"


class DatePrecision(Enum):
    Year = "y"
    Month = "M"
    Day = "d"


class InterpretedDate(object):

    def __init__(self, year, month=None, day=None, type=DateType.Exact):
        self.year = int(year)
        self.month = int(month) if month else None
        self.day = int(day) if day else None
        self.type = type

    def validate(self):
        if self.year < 0:
            raise ValueError("Year {} is negative".format(self.year))
        if self.month and (self.month < 1 or self.month > 12):
            raise ValueError("Month {} is not between 1 and 12".format(self.month))
        if self.day and (self.day < 1 or self.day > 31):
            raise ValueError("Day {} is not between 1 and 31".format(self.day))
        # Check if day is correct depending on year and month
        if self.precision == DatePrecision.Day:
            try:
                datetime(*tuple(self))
            except ValueError:
                raise ValueError("Date format is incorrect {}".format(self))

    @property
    def precision(self):
        """
        Return precision of the date depending on present elements
        """
        if self.month and self.day:
            return DatePrecision.Day
        if self.month:
            return DatePrecision.Month
        return DatePrecision.Year

    def __hash__(self):
        return hash(tuple(self))

    def __iter__(self):
        return iter((self.year, self.month, self.day))

    def __eq__(self, other):
        return tuple(self) == tuple(other) and self.type == other.type

    def __gt__(self, other):
        """
        Compare dates by year, month then day if comparison is possible
        """
        for s, o in zip(tuple(self), tuple(other)):
            if s is None or o is None:
                return
            if s == o:
                continue
            return s > o

    def __str__(self):
        return "-".join("{:02d}".format(e) for e in tuple(self) if e)


class InterpretedDateMixin(object):
    """
    Adds on-demand date parsing from a text field to InterpretedDates.
    Requires a `raw_dates` property that returns the date string.
    """

    def get_dates(self):
        from arkindex.documents.date_parser import parse_date
        if not self.raw_dates:
            return []
        return parse_date(self.raw_dates)
