import logging
from functools import wraps
from io import BytesIO

import boto3.session
from botocore.config import Config
from botocore.exceptions import BotoCoreError, ClientError
from django.conf import settings
from django.utils.functional import cached_property
from enumfields import Enum
from tenacity import retry, retry_if_exception, stop_after_delay

from arkindex.project.tools import should_verify_cert

logger = logging.getLogger(__name__)


def get_s3_resource(
        access_key_id=settings.AWS_ACCESS_KEY,
        secret_access_key=settings.AWS_SECRET_KEY,
        endpoint=settings.AWS_ENDPOINT,
        region=settings.AWS_REGION):
    session = boto3.session.Session(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
    )

    config = Config(
        region_name=region,
        signature_version="s3v4",
        s3={
            "addressing_style": "auto" if endpoint else "virtual",
        }
    )

    return session.resource(
        "s3",
        endpoint_url=endpoint,
        config=config,
        verify=should_verify_cert(endpoint),
    )


s3 = get_s3_resource()


def get_ingest_resource():
    assert settings.INGEST_S3_ENDPOINT or settings.INGEST_S3_REGION, "An endpoint or region is required"
    return get_s3_resource(
        access_key_id=settings.INGEST_S3_ACCESS_KEY,
        secret_access_key=settings.INGEST_S3_SECRET_KEY,
        endpoint=settings.INGEST_S3_ENDPOINT,
        region=settings.INGEST_S3_REGION,
    )


def requires_s3_object(func):
    """
    Decorator to handle not defined s3_object property
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.s3_object is None:
            logger.debug(
                f"Function {func.__name__} requires file to have s3_object property, skipping"
            )
            return
        return func(self, *args, **kwargs)
    return wrapper


class S3FileStatus(Enum):
    """
    S3 file validation status
    """
    Checked = "checked"
    Unchecked = "unchecked"
    Error = "error"


def _retry_delete_predicate(exception):
    """
    Retry S3FileMixin.s3_delete only for botocore errors that are not an HTTP 404
    """
    return isinstance(exception, BotoCoreError) or (
        isinstance(exception, ClientError)
        and exception.response["Error"]["Code"] != "404"
    )


class S3FileMixin(object):

    def get_s3_object(self):
        if not self.s3_bucket or not self.s3_key:
            return
        return s3.Object(self.s3_bucket, self.s3_key)

    s3_object = cached_property(get_s3_object)
    s3_region = settings.AWS_REGION

    @property
    @requires_s3_object
    def s3_put_url(self):
        return s3.meta.client.generate_presigned_url("put_object", Params={
            "Bucket": self.s3_object.bucket_name,
            "Key": self.s3_object.key,
        })

    @property
    @requires_s3_object
    def s3_url(self) -> str:
        # Handle different regions signatures
        if self.s3_region != settings.AWS_REGION:
            session = boto3.session.Session(
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
            )
            client = session.client(
                "s3",
                config=Config(signature_version="s3v4"),
                region_name=self.s3_region
            )
        else:
            client = s3.meta.client
        return client.generate_presigned_url("get_object", Params={
            "Bucket": self.s3_object.bucket_name,
            "Key": self.s3_object.key,
        })

    def exists(self):
        """
        Returns whether the file exists on S3 bucket by performing a HEAD request to S3.
        See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Object.load
        """
        try:
            self.s3_object.load()
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] != "404":
                raise
            return False

    def download(self):
        b = BytesIO()
        logger.debug("Downloading file {} from S3".format(self.s3_key))
        self.s3_object.download_fileobj(b)
        return b

    def download_to(self, path):
        logger.debug("Downloading file {} from S3".format(self.s3_key))
        self.s3_object.download_file(path)

    def check_hash(self, save=True, raise_exc=False):
        """
        Checks the MD5 hash against the hash from Amazon S3
        """
        assert self.hash, "File has no hash"
        assert self.exists(), "No file content, assert file has been correctly uploaded"
        # The hash given by Boto seems to be surrounded by double quotes
        if self.s3_object.e_tag.strip('"') == self.hash:
            self.status = S3FileStatus.Checked
        elif "-" in self.s3_object.e_tag:
            # Multipart hash: a hash of each part's hash,
            # combined with the number of parts, separated by a dash
            logger.warning("Could not check remote multipart hash {!r} against local hash {!r}".format(
                self.s3_object.e_tag,
                self.hash,
            ))
            self.status = S3FileStatus.Unchecked
        else:
            self.status = S3FileStatus.Error
        if save:
            self.save()
        if self.status == S3FileStatus.Error and raise_exc:
            raise ValueError("MD5 hashes do not match")

    @retry(
        retry=retry_if_exception(_retry_delete_predicate),
        stop=stop_after_delay(10),
        reraise=True,
    )
    def s3_delete(self):
        """
        Try deleting the S3 object.
        """
        self.s3_object.delete()
