import uuid
from unittest import expectedFailure

from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessMode, Repository, WorkerType
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, ProcessACLMixin, TrainingModelMixin
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import Model
from arkindex.users.models import Group, Right, Role, User
from arkindex.users.utils import filter_rights, get_max_level


class TestACLMixin(FixtureTestCase):

    @classmethod
    def setUpClass(cls):
        r"""
        Create user and groups with rights on a Corpus, a Repository and a Model
        We use a simple rights configuration for those tests

           User1      User2     User3      User4     User5
          /  |  \      / |     /   |         |       /
         |   |  100   10 |   100   |         |    100
        50  100   \  /   90   |    75       100    |
         |   |   Group1  |  Group2 |         |  Group3
         | Worker / \    |   /\    |         |   /  \
         |   |   80  75  | 100 50  |         |  10   100
          \  |  /      \ | /    \  |         | /       \
           Repo1     Corpus1  Corpus2      Model1     Model2
        """

        super().setUpClass()
        cls.user1 = User.objects.create_user("user1@test.test", display_name="User1")
        cls.user2 = User.objects.create_user("user2@test.test", display_name="User2")
        cls.user3 = User.objects.create_user("user3@test.test", display_name="User3")
        cls.user4 = User.objects.create_user("user4@test.test", display_name="User4")
        cls.user5 = User.objects.create_user("user5@test.test", display_name="User5")

        cls.group1 = Group.objects.create(name="Group1")
        cls.group2 = Group.objects.create(name="Group2")
        cls.group3 = Group.objects.create(name="Group3")

        cls.repo1 = Repository.objects.create(url="http://repo1")
        cls.worker_type = WorkerType.objects.create(display_name="ner", slug="ner")
        cls.worker = cls.repo1.workers.create(name="repo1 worker", slug="worker", type=cls.worker_type)

        cls.corpus1 = Corpus.objects.create(name="Corpus1", id=uuid.UUID("bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"))
        cls.corpus2 = Corpus.objects.create(name="Corpus2", id=uuid.UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"))

        cls.model1 = Model.objects.create(name="Model1", id=uuid.UUID("cccccccc-cccc-cccc-cccc-cccccccccccc"))
        cls.model2 = Model.objects.create(name="Model2", id=uuid.UUID("dddddddd-dddd-dddd-dddd-dddddddddddd"))

        Right.objects.bulk_create([
            Right(user=cls.user1, content_object=cls.group1, level=100),
            Right(user=cls.user1, content_object=cls.repo1, level=50),
            Right(user=cls.user1, content_object=cls.worker, level=100),
            Right(user=cls.user2, content_object=cls.group1, level=10),
            Right(user=cls.user2, content_object=cls.corpus1, level=90),
            Right(user=cls.user3, content_object=cls.group2, level=100),
            Right(user=cls.user3, content_object=cls.corpus2, level=75),
            Right(user=cls.user4, content_object=cls.model1, level=100),
            Right(user=cls.user5, content_object=cls.group3, level=100),
            Right(group=cls.group1, content_object=cls.repo1, level=80),
            Right(group=cls.group1, content_object=cls.corpus1, level=75),
            Right(group=cls.group2, content_object=cls.corpus1, level=100),
            Right(group=cls.group2, content_object=cls.corpus2, level=50),
            Right(group=cls.group3, content_object=cls.model1, level=10),
            Right(group=cls.group3, content_object=cls.model2, level=100),
        ])
        cls.corpus_type = ContentType.objects.get_for_model(Corpus)
        cls.group_type = ContentType.objects.get_for_model(Group)

    @expectedFailure
    def test_filter_right(self):
        # List all the rights with a privilege greater than 80 for user 3
        params = {
            "user_id": self.user3.id,
            "group_id": self.group2.id,
            "group_type_id": self.group_type.id,
            "corpus_type_id": self.corpus_type.id,
            "level": 80,
        }
        with self.assertExactQueries("corpus_rights_filter.sql", params=params):
            # List to queryset to fire the DB request
            corpora = list(filter_rights(self.user3, Corpus, 80))
        self.assertCountEqual(corpora, [self.corpus1])

    @expectedFailure
    def test_filter_right_including_public(self):
        # Listing readable objects involve listing public objects with a SQL Union
        params = {
            "user_id": self.user1.id,
            "group_id": self.group2.id,
            "group_type_id": self.group_type.id,
            "corpus_type_id": self.corpus_type.id,
            "level": Role.Guest.value,
        }
        with self.assertExactQueries("corpus_rights_filter_public.sql", params=params):
            corpora = list(filter_rights(self.user1, Corpus, Role.Guest.value))
        self.assertCountEqual(
            list(corpora),
            [self.corpus1, self.corpus]
        )

    @expectedFailure
    def test_right_direct_access(self):
        # User 2 has a direct access to the above corpus with a level of 90
        acl_mixin = ACLMixin(user=self.user2)
        params = {
            "user_id": self.user2.id,
            "group_id": self.group2.id,
            "corpus_id": self.corpus1.id,
            "group_type_id": self.group_type.id,
            "corpus_type_id": self.corpus_type.id,
            "level": 80,
        }
        with self.assertExactQueries("corpus_right_access.sql", params=params):
            has_access = acl_mixin.has_access(self.corpus1, 80)
        self.assertTrue(has_access)

    @expectedFailure
    def test_right_group_members_restriction(self):
        # User rights on corpora via a group are restricted to user level inside the group
        acl_mixin = ACLMixin(user=self.user1)
        self.assertFalse(
            acl_mixin.has_access(self.corpus1, 80)
        )

    def test_corpus_acl_mixin_has_read_access(self):
        corpus_acl_mixin = CorpusACLMixin(user=self.user2)
        with self.assertNumQueries(0):
            read_access = corpus_acl_mixin.has_read_access(self.corpus1)
        self.assertTrue(read_access)

    def test_public_corpus_has_read_access(self):
        # Any user should have a read access to public corpora
        corpus_acl_mixin = CorpusACLMixin(user=AnonymousUser())
        with self.assertNumQueries(0):
            read_access = corpus_acl_mixin.has_read_access(self.corpus)
        self.assertTrue(read_access)

    def test_corpus_acl_mixin_has_write_access(self):
        # User2 has a direct access to Corpus1 with an adequate level
        corpus_acl_mixin = CorpusACLMixin(user=self.user2)
        with self.assertNumQueries(0):
            write_access = corpus_acl_mixin.has_write_access(self.corpus1)
        self.assertTrue(write_access)

    @expectedFailure
    def test_corpus_acl_mixin_has_admin_access(self):
        # Admin access requires either to have an admin right or to be a django admin/internal user
        admin_access = [
            (self.user, 3, False),
            (self.superuser, 0, True),
            (self.user1, 3, False),
            (self.user2, 3, False),
            (self.user3, 3, True)
        ]
        for user, queries, access_check in admin_access:
            corpus_acl_mixin = CorpusACLMixin(user=user)
            with self.assertNumQueries(queries):
                admin_access = corpus_acl_mixin.has_admin_access(self.corpus1)
            self.assertEqual(admin_access, access_check)
            ContentType.objects.clear_cache()

    @expectedFailure
    def test_max_level_does_not_exists(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.corpus2), None)

    @expectedFailure
    def test_max_level_user1(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.repo1), 80)

    @expectedFailure
    def test_max_level_user2(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user2, self.repo1), 10)

    @expectedFailure
    def test_max_level_user3(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user3, self.corpus2), 75)

    @expectedFailure
    def test_max_level_public(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.corpus), Role.Guest.value)

    @expectedFailure
    def test_process_access_project(self):
        """
        A project attached to a process defines its access rights
        """
        process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user2, corpus=self.corpus1)
        with self.assertNumQueries(3):
            self.assertEqual(ProcessACLMixin(user=self.user1).process_access_level(process), 75)

    @expectedFailure
    def test_process_access_list(self):
        """
        A user can list process if they have rights on its corresponding project or repository
        """
        # Readable processes
        process_1 = Process.objects.filter(mode=ProcessMode.Workers).get()
        process_2 = Process.objects.create(mode=ProcessMode.Local, creator=self.user1)
        process_3 = Process.objects.create(mode=ProcessMode.Workers, creator=self.user2, corpus=self.corpus1)
        # Non readable processes
        Process.objects.create(mode=ProcessMode.Local, creator=self.user2)
        Process.objects.create(mode=ProcessMode.Workers, creator=self.user2, corpus=self.corpus2)

        with self.assertNumQueries(1):
            readable_process_ids = list(
                ProcessACLMixin(user=self.user1).readable_processes.values_list("id", flat=True)
            )
        self.assertCountEqual(
            readable_process_ids,
            [process_1.id, process_2.id, process_3.id]
        )

    def test_models_access_user(self):
        """
        User5 has guest access to Model1 via Group3
        """
        with self.assertNumQueries(0):
            read_access = TrainingModelMixin(user=self.user5).has_read_access(self.model1)
        self.assertTrue(read_access)

    @expectedFailure
    def test_models_no_access_user(self):
        """
        User3 has no guest access to Model1
        """
        with self.assertNumQueries(0):
            read_access = TrainingModelMixin(user=self.user3).has_read_access(self.model1)
        self.assertFalse(read_access)

    def test_models_contrib_access(self):
        """
        User4 has contributor access to Model1
        """
        with self.assertNumQueries(0):
            write_access = TrainingModelMixin(user=self.user4).has_write_access(self.model1)
        self.assertTrue(write_access)

    def test_models_admin_access_group(self):
        """
        User5 has admin access to Model2 via Group3's access rights
        """
        with self.assertNumQueries(0):
            admin_access = TrainingModelMixin(user=self.user5).has_admin_access(self.model2)
        self.assertTrue(admin_access)
