from django.contrib.postgres.search import SearchQuery, SearchVector
from django.core.management import CommandError
from django.core.management.base import BaseCommand

from arkindex.process.models import Process, ProcessMode, Repository, WorkerVersionState
from arkindex.project.argparse import UserArgument
from arkindex.training.models import Model

# List of ML results linked to a worker version that will be updated
# when migrating a worker version towards the generic one + model
# We need to know how to retrieve their corpus_id to create dedicated processes per corpus
ML_RESULTS = {
    "elements": "corpus_id",
    "transcriptions": "element__corpus_id",
    "entities": "corpus_id",
    "transcription_entities": "entity__corpus_id",
    "metadatas": "element__corpus_id",
}


def ask(phrase):
    while True:
        choice = input(phrase + " [y/n]: ")
        if choice.lower() == "y":
            return True
        if choice.lower() == "n":
            return False


def choose(instances, name_field="name", title="Pick one item", allow_skip=False):

    # Easy immediate choices
    nb = instances.count()
    if nb == 0:
        return
    if nb == 1:
        return instances.first()

    # Build internal representation
    items = instances.order_by(name_field).values_list("id", name_field)

    # Display all choices and memorize possible keys
    choices = {}
    if allow_skip:
        print("O: skip")
        choices["0"] = choices["skip"] = None
    for i, (id, name) in enumerate(items, 1):
        print(f"{i}: {id} {name}")
        choices[str((i))] = choices[str(id)] = choices[name] = id

    # Get the first valid choice
    while True:
        choice = input(f"{title}: ")
        if choice in choices:
            value = choices[choice]
            if value is None:
                return
            return instances.get(pk=choices[choice])


class Command(BaseCommand):
    help = "Migrate all workers from a repository towards a generic worker"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "repository_url",
            help="ID of the repository to migrate",
            type=str,
        )
        parser.add_argument(
            "--user",
            help="Email or ID of the admin user that will host all Worker Runs created",
            required=True,
            type=UserArgument(),
        )

    def handle(self, repository_url, user, *args, **options):

        # Get repo from its url using exact lookup, then fallback to contains
        try:
            repo = Repository.objects.get(url=repository_url)
        except Repository.DoesNotExist:
            repos = Repository.objects.filter(url__contains=repository_url)
            try:
                repo = repos.get()
            except Repository.DoesNotExist as e:
                raise CommandError(f"No repository found: {e}")
            except Repository.MultipleObjectsReturned:
                print(f"Multiple repositories found for {repository_url!r}")
                repo = choose(repos, name_field="url")

        if not repo.workers.exists():
            raise CommandError(f"No workers available in repo {repo.url}")

        # Find the generic worker
        generics = repo.workers.filter(name__icontains="generic")
        nb = generics.count()
        if nb >= 1:
            # Pick one
            generic = choose(generics, title="Pick a generic worker")
        else:
            # Pick one amongst all
            generic = choose(repo.workers, title="Pick a generic worker")

        print(f"Selected generic worker: {generic.id} {generic.name}")

        # Pick the latest available version on main/master
        generic_version = generic.versions.filter(revision__refs__name__in=("main", "master"), state=WorkerVersionState.Available).first()
        if generic_version is None:
            raise CommandError(f"No available worker version on main/master for generic worker {generic.name}")

        # Iterate over all repo's workers, except:
        # - the generic one, it's the target
        # - training workers
        workers = repo.workers.exclude(id=generic.id).exclude(name__icontains="training").order_by("name")
        for worker in workers:
            print("-" * 80)
            print(f"Migrating worker {worker.id} {worker.name}")

            # Check it has some results
            # delete version if it's unused
            for version in worker.versions.all():
                if not version.elements.exists() \
                   and not version.transcriptions.exists() \
                   and not version.entities.exists() \
                   and not version.transcription_entities.exists() \
                   and not version.metadatas.exists():
                    version.delete()
                    print(f"Deleted {version}")

            # Try to find a related model with same name
            models = Model.objects.exclude(versions__isnull=True)
            models = models.annotate(search=SearchVector("name")).filter(search=SearchQuery(worker.name, search_type="phrase"))
            model = choose(models, title=f"Pick a model to match {worker.name}", allow_skip=True)
            if model is None:
                print(f"No model found for worker {worker.name}")
                self.cleanup_worker(worker)
                continue

            # Mark this worker as compatible with the generic model
            if not model.compatible_workers.contains(generic):
                model.compatible_workers.add(generic)

            # Pick the latest model version available as we cannot match
            # a model version for a worker version
            model_version = model.versions.order_by("created").last()

            # Confirm before applying changes
            if not ask(f"Are you sure to migrate:\n - worker {worker.name}\n - towards {generic.name} + model {model.name}\n"):
                continue

            # Build worker run for that combo
            # and assign it to all ML results
            for version in worker.versions.all():
                print(f" > Migrating {version}")

                # List all corpus impacted by that worker version using all ml results
                corpus_ids = set()
                for result_key, corpus_relation in ML_RESULTS.items():
                    target = getattr(version, result_key)
                    corpus_ids |= set(target.values_list(corpus_relation, flat=True).distinct())

                for corpus_id in corpus_ids:
                    # Create a dedicated worker process by impacted corpus
                    # to migrate that worker
                    worker_run = self.build_worker_run(corpus_id, user, generic_version, model_version)

                    # Update all ML results for that worker version
                    self.update_results(version, corpus_id, worker_run)

                # Finally remove that version
                version.delete()

            # We can safely delete the full worker if no versions are remaining
            self.cleanup_worker(worker)

    def build_worker_run(self, corpus_id, user, worker_version, model_version):
        # Create a worker run on a worker process tied to a corpus
        # while respecting the constraint for unique worker version per process (no model in that constraint)
        base_name = "Migrate workers and models towards generic worker"
        worker_run = None
        part = 1
        while worker_run is None:
            suffix = f" - part {part}" if part > 1 else ""

            # Get a shared worker process on that corpus
            process, _ = Process.objects.get_or_create(
                creator=user,
                mode=ProcessMode.Workers,
                corpus_id=corpus_id,
                name=base_name + suffix,
            )

            try:
                # Create a worker run in that process when possible
                worker_run, _ = process.worker_runs.get_or_create(
                    version=worker_version,
                    model_version=model_version,
                    defaults={"parents": []},
                )
            except Exception:
                worker_run = None
                part += 1

        return worker_run

    def update_results(self, version, corpus_id, worker_run):
        # Re-assign all worker results from a worker version towards a new worker run
        for result_key, corpus_relation in ML_RESULTS.items():
            target = getattr(version, result_key)
            nb = target.filter(**{corpus_relation: corpus_id}).update(
                worker_version=worker_run.version,
                worker_run=worker_run,
            )
            print(f" > Updated {nb} {result_key}")

    def cleanup_worker(self, worker):
        # Only delete workers without any versions remaining
        if worker.versions.exists():
            return

        worker.delete()
        print(f"Deleted worker {worker.name}")
